# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Utilities that are common to multiple CLI commands."""

from collections.abc import Mapping
from dataclasses import dataclass
from functools import cached_property
from pathlib import Path
from typing import Any

import click

from ..copyright import SpdxExpression
from ..exceptions import GlobalLicensingConflictError, GlobalLicensingParseError
from ..i18n import _
from ..project import Project
from ..vcs import find_root


@dataclass()
class ClickObj:
    """A dataclass holding necessary context and options."""

    root: Path | None = None
    include_submodules: bool = False
    include_meson_subprojects: bool = False
    no_multiprocessing: bool = True

    @cached_property
    def project(self) -> Project:
        """Generate a project object on demand."""
        root = self.root
        if root is None:
            root = find_root()
        if root is None:
            root = Path.cwd()

        try:
            return Project.from_directory(
                root,
                include_submodules=self.include_submodules,
                include_meson_subprojects=self.include_meson_subprojects,
            )
        # FileNotFoundError and NotADirectoryError don't need to be caught
        # because argparse already made sure of these things.
        except GlobalLicensingParseError as error:
            raise click.UsageError(
                _(
                    "'{path}' could not be parsed. We received the"
                    " following error message: {message}"
                ).format(path=error.source, message=str(error))
            ) from error

        except (GlobalLicensingConflictError, OSError) as error:
            raise click.UsageError(str(error)) from error


class MutexOption(click.Option):
    """Enable declaring mutually exclusive options."""

    def __init__(self, *args: Any, **kwargs: Any):
        self.mutually_exclusive: set[str] = set(
            kwargs.pop("mutually_exclusive", [])
        )
        super().__init__(*args, **kwargs)
        # If self is in mutex, remove it.
        self.mutually_exclusive -= {self.name}

    @staticmethod
    def _get_long_name(ctx: click.Context, name: str) -> str:
        """Given the option name, get the long name of the option.

        For example, 'output' return '--output'.
        """
        param = next(
            param for param in ctx.command.params if param.name == name
        )
        return param.opts[0]

    def handle_parse_result(
        self, ctx: click.Context, opts: Mapping[str, Any], args: list[str]
    ) -> tuple[Any, list[str]]:
        if self.mutually_exclusive.intersection(opts) and self.name in opts:
            raise click.UsageError(
                _("'{name}' is mutually exclusive with: {opts}").format(
                    name=self._get_long_name(ctx, str(self.name)),
                    opts=", ".join(
                        f"'{self._get_long_name(ctx, opt)}'"
                        for opt in self.mutually_exclusive
                    ),
                )
            )
        return super().handle_parse_result(ctx, opts, args)


def spdx_identifier(text: str) -> SpdxExpression:
    """Factory for creating SPDX expressions."""
    expression = SpdxExpression(text)
    if not expression.is_valid:
        raise click.UsageError(
            _("'{}' is not a valid SPDX expression.").format(text)
        )
    return expression



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Some typing definitions."""

from os import PathLike
from typing import Union

#: Something that looks like a path.
StrPath = Union[str, PathLike[str]]



# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""All command-line functionality."""

from . import (
    annotate,
    convert_dep5,
    download,
    lint,
    lint_file,
    main,
    spdx,
    supported_licenses,
)



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2020 Tuomas Siipola <tuomas@zpl.fi>
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Nico Rikken <nico.rikken@fsfe.org>
# SPDX-FileCopyrightText: 2022 Pietro Albini <pietro.albini@ferrous-systems.com>
# SPDX-FileCopyrightText: 2023 DB Systel GmbH
# SPDX-FileCopyrightText: 2023 Johannes Zarl-Zierl <johannes@zarl-zierl.at>
# SPDX-FileCopyrightText: 2024 Rivos Inc.
# SPDX-FileCopyrightText: 2024 Skyler Grey <sky@a.starrysky.fyi>
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
# SPDX-FileCopyrightText: 2025 Simon Barth <simon.barth@gmx.de>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Utilities related to the extraction of REUSE information out of files."""

import codecs
import contextlib
import importlib
import logging
import os
import platform
import re
import sys
from encodings import aliases, normalize_encoding
from itertools import chain
from types import ModuleType
from typing import BinaryIO, Generator, Literal, NamedTuple, cast

from .comment import _all_style_classes
from .copyright import (
    COPYRIGHT_NOTICE_PATTERN,
    CopyrightNotice,
    ReuseInfo,
    SpdxExpression,
)
from .exceptions import NoEncodingModuleError
from .i18n import _

_LOGGER = logging.getLogger(__name__)

_ENCODING_MODULES = {  # pylint: disable=invalid-name
    "python-magic": "magic",
    "file-magic": "magic",
    "charset_normalizer": "charset_normalizer",
    "chardet": "chardet",
}
if _env_encoding_module := os.environ.get("REUSE_ENCODING_MODULE"):
    # Backwards compatibility. In v6.1.2, 'magic' used to mean 'python-magic'.
    if _env_encoding_module == "magic":
        _env_encoding_module = "python-magic"
    if _env_encoding_module not in _ENCODING_MODULES:
        print(
            # TRANSLATORS: Do not translate REUSE_ENCODING_MODULE.
            _(
                "REUSE_ENCODING_MODULE must have a value in {modules}; it has"
                " '{env_module}'. Aborting."
            ).format(
                modules=list(_ENCODING_MODULES.keys()),
                env_module=_env_encoding_module,
            )
        )
        sys.exit(1)
    _ENCODING_MODULES = {  # pylint: disable=invalid-name
        _env_encoding_module: (
            "magic" if "magic" in _env_encoding_module else _env_encoding_module
        )
    }

_ENCODING_MODULE: ModuleType | None = None
_MAGIC = None
for _module in _ENCODING_MODULES.values():
    if _module == "magic" and platform.system() == "Windows":
        continue
    try:
        _ENCODING_MODULE = importlib.import_module(_module)
        break
    except ImportError:
        continue
else:
    raise NoEncodingModuleError(
        _(
            "No supported module that can detect the encoding of files could be"
            " successfully imported. Re-read the installation instructions for"
            " the reuse package, or try the following:"
        )
        + "\n\n"
        + _(
            "- If you are running a Linux distribution, try your equivalent of"
            " `apt install file` or `dnf install file`."
        )
        + "\n"
        + _(
            "- Run ` pipx install reuse[charset-normalizer]`. Replace 'pipx'"
            " with 'pip' if you are not using pipx."
        )
    )


def _detect_magic(module: ModuleType) -> Literal["python-magic", "file-magic"]:
    if hasattr(module, "from_buffer"):
        return "python-magic"
    return "file-magic"


def get_encoding_module() -> ModuleType:
    """Get the module used to detect the encodings of files."""
    return cast(ModuleType, _ENCODING_MODULE)


def _get_encoding_module_name() -> str | None:
    module = get_encoding_module()
    result = getattr(module, "__name__", None)
    if result == "magic":
        return _detect_magic(module)
    return result


def set_encoding_module(name: str) -> ModuleType:
    """Set the module used to detect the encodings of files, and return the
    module.
    """
    if name not in _ENCODING_MODULES:
        raise NoEncodingModuleError(f"'{name}' is not a valid encoding module.")
    try:
        # pylint: disable=global-statement
        global _ENCODING_MODULE
        _ENCODING_MODULE = importlib.import_module(name)
        return _ENCODING_MODULE
    except ImportError as err:
        raise NoEncodingModuleError(f"'{name}' could not be imported.") from err


if _get_encoding_module_name() == "python-magic":
    _MAGIC = get_encoding_module().Magic(mime_encoding=True)


REUSE_IGNORE_START = "REUSE-IgnoreStart"
REUSE_IGNORE_END = "REUSE-IgnoreEnd"

# REUSE-IgnoreStart

SPDX_SNIPPET_INDICATOR = b"SPDX-SnippetBegin"

_START_PATTERN = r"(?:^.*?)"
_END_PATTERN = r"\s*(?:{})*\s*$".format(
    "|".join(
        set(
            chain(
                (
                    re.escape(style.MULTI_LINE.end)
                    for style in _all_style_classes()
                    if style.MULTI_LINE.end
                ),
                # These are special endings which do not belong to specific
                # comment styles, but which we want to nonetheless strip away
                # while parsing.
                (
                    # ex: <tag value="Copyright Jane Doe">
                    r'"\s*/*>',
                    r"'\s*/*>",
                    # ex: [SPDX-License-Identifier: GPL-3.0-or-later] ::
                    r"\]\s*::",
                ),
            )
        )
    )
)
_ALL_MATCH_PATTERN = re.compile(
    r"^(?P<prefix>.*?)(?:SPDX-\S+:|Copyright|©).*$",
    re.MULTILINE,
)
_COPYRIGHT_NOTICE_PATTERN = re.compile(
    _START_PATTERN + COPYRIGHT_NOTICE_PATTERN.pattern + _END_PATTERN,
)
_LICENSE_IDENTIFIER_PATTERN = re.compile(
    _START_PATTERN
    + r"SPDX-License-Identifier:\s*(?P<value>.*?)"
    + _END_PATTERN,
)
_CONTRIBUTOR_PATTERN = re.compile(
    _START_PATTERN + r"SPDX-FileContributor:\s*(?P<value>.*?)" + _END_PATTERN,
)
# The keys match the relevant attributes of ReuseInfo.
_SPDX_TAGS: dict[str, re.Pattern] = {
    "spdx_expressions": _LICENSE_IDENTIFIER_PATTERN,
    "contributor_lines": _CONTRIBUTOR_PATTERN,
}
_LICENSEREF_PATTERN = re.compile(r"LicenseRef-[a-zA-Z0-9-.]+$")
_NEWLINE_PATTERN = re.compile(r"\r\n?")

_LINE_ENDINGS = ("\r\n", "\r", "\n")
_LINE_ENDINGS_ASCII = tuple(ending.encode("ascii") for ending in _LINE_ENDINGS)
_LINE_ENDINGS_UTF_16_LE = tuple(
    ending.encode("utf_16_le") for ending in _LINE_ENDINGS
)
_LINE_ENDINGS_UTF_32_LE = tuple(
    ending.encode("utf_32_le") for ending in _LINE_ENDINGS
)
_LINE_ENDING_ENCODINGS_ASCII = set()
for _name in set(chain.from_iterable(aliases.aliases.items())):
    with contextlib.suppress(Exception):
        if codecs.encode("\r\n", _name) == b"\r\n":
            _LINE_ENDING_ENCODINGS_ASCII.add(_name)
_LINE_ENDING_ENCODINGS_ASCII.add("utf_8_sig")
_LINE_ENDING_ENCODINGS_UTF_16_LE = {
    "u16",
    "utf16",
    "utf_16",
    "unicodelittleunmarked",
    "utf_16le",
    "utf_16_le",
}
_LINE_ENDING_ENCODINGS_UTF_32_LE = {
    "u32",
    "utf32",
    "utf_32",
    "utf_32le",
    "utf_32_le",
}


#: Default chunk size for reading files.
CHUNK_SIZE = 1024 * 64
#: Default line size for reading files.
LINE_SIZE = 1024
#: Default chunk size used to heuristically detect file type, encoding, et
#: cetera.
HEURISTICS_CHUNK_SIZE = 1024 * 2


class FilterBlock(NamedTuple):
    """A simple tuple that holds a block of text, and whether that block of text
    is in an ignore block.
    """

    text: str
    in_ignore_block: bool


def filter_ignore_block(
    text: str, in_ignore_block: bool = False
) -> FilterBlock:
    """Filter out blocks beginning with REUSE_IGNORE_START and ending with
    REUSE_IGNORE_END to remove lines that should not be treated as copyright and
    licensing information.

    Args:
        text: The text out of which the ignore blocks must be filtered.
        in_ignore_block: Whether the text is already in an ignore block. This is
            useful when you parse subsequent chunks of text, and one chunk does
            not close the ignore block.

    Returns:
        A :class:`FilterBlock` tuple that contains the filtered text and a
        boolean that signals whether the ignore block is still open.
    """
    ignore_start: int | None = None if not in_ignore_block else 0
    ignore_end: int | None = None
    if REUSE_IGNORE_START in text:
        ignore_start = text.index(REUSE_IGNORE_START)
    if REUSE_IGNORE_END in text:
        ignore_end = text.index(REUSE_IGNORE_END) + len(REUSE_IGNORE_END)
    if ignore_start is None:
        return FilterBlock(text, False)
    if ignore_end is None:
        return FilterBlock(text[:ignore_start], True)
    if ignore_start < ignore_end:
        text_before_block = text[:ignore_start]
        text_after_block, in_ignore_block = filter_ignore_block(
            text[ignore_end:], False
        )
        return FilterBlock(
            text_before_block + text_after_block, in_ignore_block
        )
    rest = text[ignore_start + len(REUSE_IGNORE_START) :]
    if REUSE_IGNORE_END in rest:
        ignore_end = rest.index(REUSE_IGNORE_END) + len(REUSE_IGNORE_END)
        text_before_block = text[:ignore_start]
        text_after_block, in_ignore_block = filter_ignore_block(
            rest[ignore_end:]
        )
        return FilterBlock(
            text_before_block + text_after_block, in_ignore_block
        )
    return FilterBlock(text[:ignore_start], True)


def extract_reuse_info(text: str) -> ReuseInfo:
    """Extract REUSE information from a multi-line text block.

    Raises:
        ExpressionError: if an SPDX expression could not be parsed.
        ParseError: if an SPDX expression could not be parsed.
    """
    notices: set[CopyrightNotice] = set()
    expressions: set[SpdxExpression] = set()
    contributors: set[str] = set()

    for possible in _ALL_MATCH_PATTERN.finditer(text):
        possible_text = possible.group()
        prefix = possible.group("prefix").strip()
        reversed_prefix = prefix[::-1]
        possible_text = possible_text.removesuffix(reversed_prefix)
        if match := _COPYRIGHT_NOTICE_PATTERN.match(possible_text):
            notices.add(CopyrightNotice.from_match(match))
        elif match := _LICENSE_IDENTIFIER_PATTERN.match(possible_text):
            expressions.add(SpdxExpression(match.group("value")))
        elif match := _CONTRIBUTOR_PATTERN.match(possible_text):
            contributors.add(match.group("value"))

    return ReuseInfo(
        spdx_expressions=expressions,
        copyright_notices=notices,
        contributor_lines=contributors,
    )


def _read_chunks(
    fp: BinaryIO,
    chunk_size: int = CHUNK_SIZE,
    line_size: int = LINE_SIZE,
    newline: bytes = b"\n",
) -> Generator[bytes, None, None]:
    """Read and yield somewhat equal-sized chunks from (realistically) a file.
    The chunks always split at a newline where possible.

    An amount of bytes equal to *chunk_size* is always read into the chunk if
    *fp* contains that many bytes. An additional *line_size* or lesser amount of
    bytes is also read into the chunk, up to the next newline character.

    *newline* is the line separator that is (expected to be) used in the input.
    """
    newline_len = len(newline)
    while True:
        chunk = fp.read(chunk_size)
        if not chunk:
            break
        end_chunk_pos = fp.tell()
        remainder = fp.read(line_size)
        newline_idx = remainder.find(newline)
        if newline_idx != -1:
            remainder = remainder[: newline_idx + newline_len]
            fp.seek(end_chunk_pos + newline_idx + newline_len)
        chunk += remainder
        yield chunk


def _detect_encoding_magic(mime_encoding: str, chunk: bytes) -> str | None:
    if mime_encoding == "binary":
        return None
    if mime_encoding == "utf-8" and chunk[:3] == b"\xef\xbb\xbf":
        mime_encoding += "-sig"
    # Python and magic disagree on what 'le' means. For magic, it means a UTF-16
    # block prefixed with a BOM. For Python, that's what 'utf-16' is, and
    # 'utf_16_le' is a little endian block _without_ BOM prefix.
    elif mime_encoding == "utf-16le" and chunk[:2] == b"\xff\xfe":
        mime_encoding = "utf-16"
    elif mime_encoding == "utf-32le" and chunk[:4] == b"\xff\xfe\x00\x00":
        mime_encoding = "utf-32"
    else:
        # This nifty function gets the (in Python) standardised name for an
        # encoding. 'iso-8859-1' becomes 'iso8859-1'.
        try:
            codec_info = codecs.lookup(mime_encoding)
            mime_encoding = codec_info.name
        except LookupError:
            # Fallback.
            mime_encoding = "utf-8"
    return normalize_encoding(mime_encoding)


def _detect_encoding_python_magic(chunk: bytes) -> str | None:
    result: str = _MAGIC.from_buffer(chunk)  # type: ignore[union-attr]
    return _detect_encoding_magic(result, chunk)


def _detect_encoding_file_magic(chunk: bytes) -> str | None:
    result: str = get_encoding_module().detect_from_content(chunk).encoding
    return _detect_encoding_magic(result, chunk)


def _detect_encoding_charset_normalizer(chunk: bytes) -> str | None:
    matches = get_encoding_module().from_bytes(  # type: ignore[union-attr]
        chunk,
    )
    best = matches.best()
    if best is not None:
        result: str = best.encoding
        if result == "utf_8" and best.bom:
            result += "_sig"
        return result
    return None


def _detect_encoding_chardet(chunk: bytes) -> str | None:
    dict_ = get_encoding_module().detect(chunk)  # type: ignore[union-attr]
    result: str | None = dict_.get("encoding")
    if result is None:
        return None
    try:
        codec_info = codecs.lookup(result)
        result = codec_info.name
    except LookupError:
        # Fallback.
        result = "utf-8"
    return normalize_encoding(result)


def detect_encoding(chunk: bytes) -> str | None:
    """Find the encoding of the bytes chunk, and return it as normalised name.
    See :func:`encodings.normalize_encoding`. If no encoding could be found,
    return :const:`None`.

    If the chunk is empty or the encoding of the chunk is ASCII, ``'utf_8'`` is
    returned.
    """
    # If the file is empty, assume UTF-8.
    if not chunk:
        return "utf_8"

    result: str | None = None
    if _get_encoding_module_name() == "python-magic":
        result = _detect_encoding_python_magic(chunk)
    elif _get_encoding_module_name() == "file-magic":
        result = _detect_encoding_file_magic(chunk)
    elif _get_encoding_module_name() == "charset_normalizer":
        result = _detect_encoding_charset_normalizer(chunk)
    elif _get_encoding_module_name() == "chardet":
        result = _detect_encoding_chardet(chunk)
    else:
        # This code should technically never be reached.
        raise NoEncodingModuleError()

    if result in ["ascii", "us_ascii"]:
        result = "utf_8"
    return result


def detect_newline(chunk: bytes, encoding: str = "ascii") -> str:
    """Return one of ``'\\n'``, ``'\\r'`` or ``'\\r\\n'`` depending on the line
    endings used in *chunk*. Return :const:`os.linesep` if there are no line
    endings.
    """
    line_endings: tuple[bytes, ...] | None = None
    encoding = normalize_encoding(encoding.lower())
    # This step is part optimalisation, part dealing with BOMs.
    if encoding in _LINE_ENDING_ENCODINGS_ASCII:
        line_endings = _LINE_ENDINGS_ASCII
    elif encoding in _LINE_ENDING_ENCODINGS_UTF_16_LE:
        line_endings = _LINE_ENDINGS_UTF_16_LE
    elif encoding in _LINE_ENDING_ENCODINGS_UTF_32_LE:
        line_endings = _LINE_ENDINGS_UTF_32_LE

    if line_endings is not None:
        for line_ending_bytes in line_endings:
            if line_ending_bytes in chunk:
                return line_ending_bytes.decode(encoding)
    else:
        for line_ending_str in _LINE_ENDINGS:
            if line_ending_str.encode(encoding) in chunk:
                return line_ending_str
    return os.linesep


def reuse_info_of_file(
    fp: BinaryIO,
    chunk_size: int = CHUNK_SIZE,
    line_size: int = LINE_SIZE,
) -> ReuseInfo:
    """Read from *fp* to extract REUSE information. It is read in chunks of
    *chunk_size*, additionally reading up to *line_size* until the next newline.

    This function decodes the binary data into UTF-8 and removes REUSE ignore
    blocks before attempting to extract the REUSE information.
    """
    position = fp.tell()
    heuristics_chunk = fp.read(HEURISTICS_CHUNK_SIZE)
    fp.seek(position)  # Reset position.
    encoding = detect_encoding(heuristics_chunk)
    filename = getattr(fp, "name", None)
    if encoding is None:
        if filename:
            _LOGGER.info(
                _(
                    "'{path}' was detected as a binary file; not searching its"
                    " contents for REUSE information."
                ).format(path=filename)
            )
        return ReuseInfo()

    newline = detect_newline(heuristics_chunk, encoding=encoding)

    if filename:
        _LOGGER.debug(
            _(
                "extracting REUSE information from '{path}'"
                " (encoding {encoding}, encoding module {module},"
                " newline {newline})"
            ).format(
                path=filename,
                encoding=repr(encoding),
                module=repr(_get_encoding_module_name()),
                newline=repr(newline),
            )
        )
    in_ignore_block = False
    reuse_infos: list[ReuseInfo] = []
    for chunk in _read_chunks(
        fp,
        chunk_size=chunk_size,
        line_size=line_size,
        newline=newline.encode(encoding),
    ):
        text = chunk.decode(encoding, errors="replace")
        text = _NEWLINE_PATTERN.sub("\n", text)
        text, in_ignore_block = filter_ignore_block(text, in_ignore_block)
        reuse_infos.append(extract_reuse_info(text))
    return ReuseInfo().union(*reuse_infos)


def contains_reuse_info(text: str) -> bool:
    """The text contains REUSE info."""
    return bool(extract_reuse_info(filter_ignore_block(text).text))


# REUSE-IgnoreEnd



# SPDX-FileCopyrightText: 2024 Kerry McAdams <github@klmcadams>
# SPDX-FileCopyrightText: 2024 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Click code for lint-file subcommand."""

# pylint: disable=unused-argument

import sys
from collections.abc import Collection
from pathlib import Path

import click

from ..i18n import _
from ..lint import format_lines_subset
from ..report import ProjectSubsetReport
from .common import ClickObj, MutexOption
from .main import main

_OUTPUT_MUTEX = ["quiet", "lines"]

_HELP = _(
    "Lint individual files for REUSE compliance. The specified FILEs are"
    " checked for the presence of copyright and licensing information, and"
    " whether the found licenses are included in the LICENSES/ directory."
)


@main.command(name="lint-file", help=_HELP)
@click.option(
    "--quiet",
    "-q",
    cls=MutexOption,
    mutually_exclusive=_OUTPUT_MUTEX,
    is_flag=True,
    help=_("Prevent output."),
)
@click.option(
    "--lines",
    "-l",
    cls=MutexOption,
    mutually_exclusive=_OUTPUT_MUTEX,
    is_flag=True,
    help=_("Format output as errors per line. (default)"),
)
@click.argument(
    "files",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("FILE"),
    type=click.Path(exists=True, path_type=Path),
    nargs=-1,
)
@click.pass_obj
def lint_file(
    obj: ClickObj, quiet: bool, lines: bool, files: Collection[Path]
) -> None:
    # pylint: disable=missing-function-docstring
    project = obj.project
    subset_files = {Path(file_) for file_ in files}
    for file_ in subset_files:
        if not file_.resolve().is_relative_to(project.root.resolve()):
            raise click.UsageError(
                _("'{file}' is not inside of '{root}'.").format(
                    file=file_, root=project.root
                )
            )
    report = ProjectSubsetReport.generate(
        project,
        subset_files,
        multiprocessing=not obj.no_multiprocessing,
    )

    if quiet:
        pass
    else:
        click.echo(format_lines_subset(report), nl=False)

    sys.exit(0 if report.is_compliant else 1)



# SPDX-FileCopyrightText: 2023 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Code for parsing and validating REUSE.toml."""

# mypy: disable-error-code=attr-defined

import functools
import logging
import re
from abc import ABC, abstractmethod
from collections import defaultdict
from collections.abc import Callable, Collection, Generator, Iterable
from enum import Enum
from pathlib import Path, PurePath
from typing import Any, TypeVar, cast

import attrs
import tomlkit
from attr.validators import _InstanceOfValidator as _AttrInstanceOfValidator
from debian.copyright import Copyright
from debian.copyright import Error as DebianError

from .copyright import CopyrightNotice, ReuseInfo, SourceType, SpdxExpression
from .covered_files import is_path_ignored
from .exceptions import (
    CopyrightNoticeParseError,
    GlobalLicensingParseError,
    GlobalLicensingParseTypeError,
    GlobalLicensingParseValueError,
)
from .i18n import _
from .types import StrPath
from .vcs import VCSStrategy

_LOGGER = logging.getLogger(__name__)

_T = TypeVar("_T")

#: Current version of REUSE.toml.
REUSE_TOML_VERSION = 1

#: Relation between Python attribute names and TOML keys.
_TOML_KEYS = {
    "paths": "path",
    "precedence": "precedence",
    "_copyright_notices": "SPDX-FileCopyrightText",
    "_spdx_expressions": "SPDX-License-Identifier",
}


class PrecedenceType(Enum):
    """An enum of behaviours surrounding order of precedence for entries in a
    :class:`GlobalLicensing`.
    """

    #: Aggregate the results from the file with the results from the global
    #: licensing file.
    AGGREGATE = "aggregate"
    #: Use the results that are closest to the covered file. This is typically
    #: the file itself, or the global licensing file if no REUSE information
    #: exists inside of the file.
    CLOSEST = "closest"
    #: Only use the results from the global licensing file.
    OVERRIDE = "override"


@attrs.define
class _CollectionOfValidator:
    collection_type: type[Collection] = attrs.field()
    value_type: type = attrs.field()
    optional: bool = attrs.field(default=True)

    def __call__(
        self,
        instance: object,
        attribute: attrs.Attribute,
        value: Collection[_T],
    ) -> None:
        # This is a hack to display the TOML's key names instead of the Python
        # attributes.
        if isinstance(instance, AnnotationsItem):
            attr_name = _TOML_KEYS[attribute.name]
        else:
            attr_name = attribute.name
        source = getattr(instance, "source", None)

        if not isinstance(value, self.collection_type):
            raise GlobalLicensingParseTypeError(
                _(
                    "{attr_name} must be a {type_name} (got {value} that is a"
                    " {value_class})."
                ).format(
                    attr_name=repr(attr_name),
                    type_name=self.collection_type.__name__,
                    value=repr(value),
                    value_class=repr(value.__class__),
                ),
                source=source,
            )
        for item in value:
            if not isinstance(item, self.value_type):
                raise GlobalLicensingParseTypeError(
                    _(
                        "Item in {attr_name} collection must be a {type_name}"
                        " (got {item_value} that is a {item_class})."
                    ).format(
                        attr_name=repr(attr_name),
                        type_name=self.value_type.__name__,
                        item_value=repr(item),
                        item_class=repr(item.__class__),
                    ),
                    source=source,
                )
        if not self.optional and not value:
            raise GlobalLicensingParseValueError(
                _("{attr_name} must not be empty.").format(
                    attr_name=repr(attr_name),
                ),
                source=source,
            )


def _validate_collection_of(
    collection_type: type[Collection],
    value_type: type[_T],
    optional: bool = False,
) -> Callable[[Any, attrs.Attribute, Collection[_T]], Any]:
    return _CollectionOfValidator(
        collection_type, value_type, optional=optional
    )


class _InstanceOfValidator(_AttrInstanceOfValidator):
    def __call__(self, inst: Any, attr: attrs.Attribute, value: _T) -> None:
        try:
            super().__call__(inst, attr, value)
        except TypeError as error:
            raise GlobalLicensingParseTypeError(
                _(
                    "{name} must be a {type} (got {value} that is a"
                    " {value_type})."
                ).format(
                    name=repr(error.args[1].name),
                    type=repr(error.args[2].__name__),
                    value=repr(error.args[3]),
                    value_type=repr(error.args[3].__class__),
                ),
                source=getattr(inst, "source", None),
            ) from error


def _instance_of(
    type_: type[_T],
) -> Callable[[Any, attrs.Attribute, _T], Any]:
    return _InstanceOfValidator(type_)


def _str_to_global_precedence(value: Any) -> PrecedenceType:
    try:
        return PrecedenceType(value)
    except ValueError as error:
        raise GlobalLicensingParseValueError(
            _(
                "The value of 'precedence' must be one of {precedence_vals}"
                " (got {received})"
            ).format(
                precedence_vals=tuple(
                    member.value for member in PrecedenceType
                ),
                received=repr(value),
            )
        ) from error


def _to_set(value: _T | Iterable[_T] | None) -> set[_T]:
    if value is None:
        return set()
    # Special case for strings.
    if isinstance(value, str):
        return {cast(_T, value)}
    if hasattr(value, "__iter__"):
        return set(value)
    return {value}


# The attrs library infers __init__ parameter types from the converter's
# signature. The signature of _to_set confuses mypy, so this wrapper exposes a
# simpler signature just for use with attrs.
def _to_set_any(value: Any | None) -> set[Any]:
    return cast(set[Any], _to_set(value))


def _to_set_of_expr(
    value: str | Iterable[str] | None,
) -> set[SpdxExpression]:
    value = _to_set(value)
    return {SpdxExpression(expression) for expression in value}


def _to_set_of_notice(
    value: str | Iterable[str] | None,
) -> set[CopyrightNotice]:
    value = _to_set(value)
    result = set()
    for notice in value:
        try:
            result.add(CopyrightNotice.from_string(notice))
        except CopyrightNoticeParseError as error:
            try:
                result.add(
                    CopyrightNotice.from_string(
                        f"SPDX-FileCopyrightText: {notice}"
                    )
                )
            except CopyrightNoticeParseError:
                raise GlobalLicensingParseValueError(
                    _("Could not parse '{notice}'").format(notice=notice)
                ) from error
    return result


@attrs.define(frozen=True)
class GlobalLicensing(ABC):
    """An abstract class that represents a configuration file that contains
    licensing information that is pertinent to other files in the project.
    """

    source: str = attrs.field(validator=_instance_of(str))

    @classmethod
    @abstractmethod
    def from_file(cls, path: StrPath, **kwargs: Any) -> "GlobalLicensing":
        """Parse the file and create a :class:`GlobalLicensing` object from its
        contents.

        Raises:
            FileNotFoundError: file doesn't exist.
            OSError: some other error surrounding I/O.
            GlobalLicensingParseError: file could not be parsed.
        """

    @abstractmethod
    def reuse_info_of(
        self, path: StrPath
    ) -> dict[PrecedenceType, list[ReuseInfo]]:
        """Find the REUSE information of *path* defined in the configuration.
        The path must be relative to the root of a
        :class:`reuse.project.Project`.

        The key indicates the precedence type for the subsequent information.
        """


@attrs.define(frozen=True)
class ReuseDep5(GlobalLicensing):
    """A soft wrapper around :class:`Copyright`."""

    dep5_copyright: Copyright

    @classmethod
    def from_file(cls, path: StrPath, **kwargs: Any) -> "ReuseDep5":
        path = Path(path)
        try:
            with path.open(encoding="utf-8") as fp:
                return cls(str(path), Copyright(fp))
        except UnicodeDecodeError as error:
            raise GlobalLicensingParseError(
                str(error), source=str(path)
            ) from error
        # TODO: Remove ValueError once
        # <https://salsa.debian.org/python-debian-team/python-debian/-/merge_requests/123>
        # is closed
        except (DebianError, ValueError) as error:
            raise GlobalLicensingParseError(
                str(error), source=str(path)
            ) from error

    def reuse_info_of(
        self, path: StrPath
    ) -> dict[PrecedenceType, list[ReuseInfo]]:
        path = PurePath(path).as_posix()
        result = self.dep5_copyright.find_files_paragraph(path)

        if result is None:
            return {}

        return {
            PrecedenceType.AGGREGATE: [
                ReuseInfo(
                    spdx_expressions=_to_set_of_expr(result.license.synopsis),
                    copyright_notices=_to_set_of_notice(
                        map(str.strip, result.copyright.splitlines())
                    ),
                    path=path,
                    source_type=SourceType.DEP5,
                    # This is hardcoded. It must be a relative path from the
                    # project root. self.source is not (guaranteed) a relative
                    # path.
                    source_path=".reuse/dep5",
                )
            ]
        }


@attrs.define(frozen=True)
class AnnotationsItem:
    """A class that maps to a single [[annotations]] table element in
    REUSE.toml.
    """

    paths: set[str] = attrs.field(
        converter=_to_set_any,
        validator=_validate_collection_of(set, str, optional=False),
    )
    precedence: PrecedenceType = attrs.field(
        converter=_str_to_global_precedence, default=PrecedenceType.CLOSEST
    )
    _copyright_notices: set[str] = attrs.field(
        alias="copyright_notices",
        converter=_to_set_any,
        validator=_validate_collection_of(set, str, optional=True),
        default=None,
    )
    _spdx_expressions: set[str] = attrs.field(
        alias="spdx_expressions",
        converter=_to_set_any,
        validator=_validate_collection_of(set, str, optional=True),
        default=None,
    )

    def __attrs_post_init__(self) -> None:
        # Immediately trigger cached properties to get error as needed.
        _ = self.copyright_notices
        _ = self.spdx_expressions

    @functools.cached_property
    def copyright_notices(self) -> set[CopyrightNotice]:
        return _to_set_of_notice(self._copyright_notices)

    @functools.cached_property
    def spdx_expressions(self) -> set[SpdxExpression]:
        return _to_set_of_expr(self._spdx_expressions)

    @functools.cached_property
    def _paths_regex(self) -> re.Pattern:
        def translate(path: str) -> str:
            # pylint: disable=too-many-branches
            blocks = []
            escaping = False
            globstar = False
            prev_char = ""
            for char in path:
                if char == "\\":
                    if prev_char == "\\" and escaping:
                        escaping = False
                        blocks.append("\\\\")
                    else:
                        escaping = True
                elif char == "*":
                    if escaping:
                        blocks.append(re.escape("*"))
                        escaping = False
                    elif prev_char == "*" and not globstar:
                        globstar = True
                        blocks.append(r".*")
                elif char == "/":
                    if not globstar:
                        if prev_char == "*":
                            blocks.append("[^/]*")
                        blocks.append("/")
                    escaping = False
                else:
                    if prev_char == "*" and not globstar:
                        blocks.append(r"[^/]*")
                    blocks.append(re.escape(char))
                    globstar = False
                    escaping = False
                prev_char = char
            if prev_char == "*" and not globstar:
                blocks.append(r"[^/]*")
            result = "".join(blocks)
            return f"^({result})$"

        return re.compile("|".join(translate(path) for path in self.paths))

    @classmethod
    def from_dict(cls, values: dict[str, Any]) -> "AnnotationsItem":
        """Create an :class:`AnnotationsItem` from a dictionary that uses the
        key-value pairs for an [[annotations]] table in REUSE.toml.
        """
        new_dict = {}
        new_dict["paths"] = values.get(_TOML_KEYS["paths"])
        precedence = values.get(_TOML_KEYS["precedence"])
        if precedence is not None:
            new_dict["precedence"] = precedence
        new_dict["copyright_notices"] = values.get(
            _TOML_KEYS["_copyright_notices"]
        )
        new_dict["spdx_expressions"] = values.get(
            _TOML_KEYS["_spdx_expressions"]
        )
        return cls(**new_dict)  # type: ignore

    def matches(self, path: str) -> bool:
        """Determine whether *path* matches any of the paths (or path globs) in
        :class:`AnnotationsItem`.
        """
        return bool(self._paths_regex.match(path))


@attrs.define(frozen=True)
class ReuseTOML(GlobalLicensing):
    """A class that contains the data parsed from a REUSE.toml file."""

    version: int = attrs.field(validator=_instance_of(int))
    annotations: list[AnnotationsItem] = attrs.field(
        validator=_validate_collection_of(list, AnnotationsItem, optional=True)
    )

    @classmethod
    def from_dict(cls, values: dict[str, Any], source: str) -> "ReuseTOML":
        """Create a :class:`ReuseTOML` from the dict version of REUSE.toml."""
        new_dict = {}
        new_dict["version"] = values.get("version")
        new_dict["source"] = source

        annotation_dicts = values.get("annotations", [])
        try:
            annotations = [
                AnnotationsItem.from_dict(annotation)
                for annotation in annotation_dicts
            ]
        except GlobalLicensingParseError as error:
            error.source = source
            raise error from error

        new_dict["annotations"] = annotations

        return cls(**new_dict)  # type: ignore

    @classmethod
    def from_toml(cls, toml: str, source: str) -> "ReuseTOML":
        """Create a :class:`ReuseTOML` from TOML text."""
        try:
            tomldict = tomlkit.loads(toml)
        except tomlkit.exceptions.TOMLKitError as error:
            raise GlobalLicensingParseError(
                str(error), source=source
            ) from error
        return cls.from_dict(tomldict, source)

    @classmethod
    def from_file(cls, path: StrPath, **kwargs: Any) -> "ReuseTOML":
        try:
            with Path(path).open(encoding="utf-8") as fp:
                return cls.from_toml(fp.read(), str(path))
        except UnicodeDecodeError as error:
            raise GlobalLicensingParseError(
                str(error), source=str(path)
            ) from error

    def find_annotations_item(self, path: StrPath) -> AnnotationsItem | None:
        """Find a :class:`AnnotationsItem` that matches *path*. The latest match
        in :attr:`annotations` is returned.
        """
        path = PurePath(path).as_posix()
        for item in reversed(self.annotations):
            if item.matches(path):
                return item
        return None

    def reuse_info_of(
        self, path: StrPath
    ) -> dict[PrecedenceType, list[ReuseInfo]]:
        path = PurePath(path).as_posix()
        item = self.find_annotations_item(path)
        if item:
            return {
                item.precedence: [
                    ReuseInfo(
                        spdx_expressions=item.spdx_expressions,
                        copyright_notices=item.copyright_notices,
                        path=path,
                        source_path="REUSE.toml",
                        source_type=SourceType.REUSE_TOML,
                    )
                ]
            }
        return {}

    @property
    def directory(self) -> PurePath:
        """The directory in which the REUSE.toml file is located."""
        return PurePath(self.source).parent


@attrs.define(frozen=True)
class NestedReuseTOML(GlobalLicensing):
    """A class that represents a hierarchy of :class:`ReuseTOML` objects."""

    reuse_tomls: list[ReuseTOML] = attrs.field()

    @classmethod
    def from_file(cls, path: StrPath, **kwargs: Any) -> "NestedReuseTOML":
        """TODO: *path* is a directory instead of a file."""
        include_submodules: bool = kwargs.get("include_submodules", False)
        include_meson_subprojects: bool = kwargs.get(
            "include_meson_subprojects", False
        )
        vcs_strategy: VCSStrategy | None = kwargs.get("vcs_strategy")
        tomls = [
            ReuseTOML.from_file(toml_path)
            for toml_path in cls.find_reuse_tomls(
                path,
                include_submodules=include_submodules,
                include_meson_subprojects=include_meson_subprojects,
                vcs_strategy=vcs_strategy,
            )
        ]
        return cls(reuse_tomls=tomls, source=str(path))

    def reuse_info_of(
        self, path: StrPath
    ) -> dict[PrecedenceType, list[ReuseInfo]]:
        path = PurePath(path)

        toml_items: list[tuple[ReuseTOML, AnnotationsItem]] = (
            self._find_relevant_tomls_and_items(path)
        )

        result = defaultdict(list)
        for keyval in toml_items:
            toml = keyval[0]
            item = keyval[1]
            relpath = (PurePath(self.source) / path).relative_to(toml.directory)
            # I'm pretty sure there should be no KeyError here.
            info = toml.reuse_info_of(relpath)[item.precedence][0]
            result[item.precedence].append(
                # Fix the paths to be relative to self.source. As-is, they
                # were relative to the directory of the respective
                # REUSE.toml.
                info.copy(
                    path=path.as_posix(),
                    source_path=PurePath(toml.source)
                    .relative_to(self.source)
                    .as_posix(),
                )
            )
            if item.precedence == PrecedenceType.OVERRIDE:
                # No more!
                break

        # Clean up CLOSEST. Some items were added that are not the closest.
        # Consider copyright and licensing separately.
        copyright_found = False
        licence_found = False
        to_keep: list[ReuseInfo] = []
        for info in reversed(result[PrecedenceType.CLOSEST]):
            new_info = info.copy(
                copyright_notices=set(), spdx_expressions=set()
            )
            if not copyright_found and info.copyright_notices:
                new_info = new_info.copy(
                    copyright_notices=info.copyright_notices
                )
                copyright_found = True
            if not licence_found and info.spdx_expressions:
                new_info = new_info.copy(spdx_expressions=info.spdx_expressions)
                licence_found = True
            if new_info.contains_copyright_or_licensing():
                to_keep.append(new_info)
        result[PrecedenceType.CLOSEST] = list(reversed(to_keep))
        # Looping over CLOSEST created it in the defaultdict. Remove it if it's
        # empty.
        if not result[PrecedenceType.CLOSEST]:
            del result[PrecedenceType.CLOSEST]

        return dict(result)

    @classmethod
    def find_reuse_tomls(
        cls,
        path: StrPath,
        include_submodules: bool = False,
        include_meson_subprojects: bool = False,
        vcs_strategy: VCSStrategy | None = None,
    ) -> Generator[Path, None, None]:
        """Find all REUSE.toml files in *path*. *path* should be the root of the
        directory. If it is not, REUSE.toml files which are in ignored
        directories may not be correctly ignored.
        """
        path = Path(path)
        reuse_tomls = path.rglob("REUSE.toml")
        for item in reuse_tomls:
            if is_path_ignored(
                item,
                include_submodules=include_submodules,
                include_meson_subprojects=include_meson_subprojects,
                include_reuse_tomls=True,
                vcs_strategy=vcs_strategy,
            ):
                continue
            rel = item.relative_to(path)
            parts = rel.parts
            for directory in (
                path.joinpath(*parts[:i]) for i in range(1, len(parts))
            ):
                if is_path_ignored(
                    directory,
                    include_submodules=include_submodules,
                    include_meson_subprojects=include_meson_subprojects,
                    vcs_strategy=vcs_strategy,
                ):
                    break
            else:
                yield item

    def _find_relevant_tomls(self, path: StrPath) -> list[ReuseTOML]:
        found = []
        for toml in self.reuse_tomls:
            if PurePath(path).is_relative_to(toml.directory):
                found.append(toml)
        # Sort from topmost to deepest directory.
        found.sort(key=lambda toml: toml.directory.parts)
        return found

    def _find_relevant_tomls_and_items(
        self, path: StrPath
    ) -> list[tuple[ReuseTOML, AnnotationsItem]]:
        # *path* is relative to the Project root, which is the *source* of
        # NestedReuseTOML, which itself is a relative (to CWD) or absolute
        # path.
        path = PurePath(path)
        adjusted_path = PurePath(self.source) / path

        tomls = self._find_relevant_tomls(adjusted_path)
        toml_items: list[tuple[ReuseTOML, AnnotationsItem]] = []
        for toml in tomls:
            relpath = adjusted_path.relative_to(toml.directory)
            item = toml.find_annotations_item(relpath)
            if item is not None:
                toml_items.append((toml, item))
        return toml_items



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2020 John Mulligan <jmulligan@redhat.com>
# SPDX-FileCopyrightText: 2023 Markus Haug <korrat@proton.me>
# SPDX-FileCopyrightText: 2024 Skyler Grey <sky@a.starrysky.fyi>
# SPDX-FileCopyrightText: 2025 Jonas Fierlings <fnoegip@gmail.com>
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""This module deals with version control systems."""

from __future__ import annotations

import logging
import os
import shutil
from abc import ABC, abstractmethod
from collections.abc import Generator
from inspect import isclass
from pathlib import Path
from typing import TYPE_CHECKING

from ._util import execute_command, relative_from_root
from .types import StrPath

if TYPE_CHECKING:
    from .project import Project

_LOGGER = logging.getLogger(__name__)

GIT_EXE = shutil.which("git")
HG_EXE = shutil.which("hg")
JUJUTSU_EXE = shutil.which("jj")
PIJUL_EXE = shutil.which("pijul")


def _find_ancestor(
    directory: StrPath, ancestor: str, is_directory: bool = True
) -> Path | None:
    path = Path(directory).resolve()
    for parent in [path] + list(path.parents):
        if (parent / ancestor).is_dir() or (
            (parent / ancestor).exists() and not is_directory
        ):
            return parent / ancestor
    return None


class VCSStrategy(ABC):
    """Strategy pattern for version control systems."""

    EXE: str | None = None

    def __init__(self, root: StrPath):
        self.root = Path(root)

    @abstractmethod
    def is_ignored(self, path: Path) -> bool:
        """Is *path* ignored by the VCS?"""

    @abstractmethod
    def is_submodule(self, path: StrPath) -> bool:
        """Is *path* a VCS submodule?"""

    @classmethod
    @abstractmethod
    def in_repo(cls, directory: StrPath) -> bool:
        """Is *directory* inside of the VCS repository?

        Raises:
            NotADirectoryError: if directory is not a directory.
        """

    @classmethod
    @abstractmethod
    def find_root(cls, cwd: StrPath | None = None) -> Path | None:
        """Try to find the root of the project from *cwd*. If none is found,
        return None.

        Raises:
            NotADirectoryError: if directory is not a directory.
        """


class VCSStrategyNone(VCSStrategy):
    """Strategy that is used when there is no VCS."""

    def is_ignored(self, path: Path) -> bool:
        return False

    def is_submodule(self, path: StrPath) -> bool:
        return False

    @classmethod
    def in_repo(cls, directory: StrPath) -> bool:
        return False

    @classmethod
    def find_root(cls, cwd: StrPath | None = None) -> Path | None:
        return None


class VCSStrategyGit(VCSStrategy):
    """Strategy that is used for Git."""

    EXE = GIT_EXE

    def __init__(self, root: StrPath):
        super().__init__(root)
        if not self.EXE:
            raise FileNotFoundError("Could not find binary for Git")
        self._all_ignored_files = self._find_all_ignored_files()
        self._submodules = self._find_submodules()

    def _find_all_ignored_files(self) -> set[Path]:
        """Return a set of all files ignored by git. If a whole directory is
        ignored, don't return all files inside of it.
        """
        command = [
            str(self.EXE),
            "ls-files",
            "--exclude-standard",
            "--ignored",
            "--others",
            "--directory",
            # TODO: This flag is unexpected.  I reported it as a bug in Git.
            # This flag---counter-intuitively---lists untracked directories
            # that contain ignored files.
            "--no-empty-directory",
            # Separate output with \0 instead of \n.
            "-z",
        ]
        result = execute_command(command, _LOGGER, cwd=self.root)
        all_files = result.stdout.decode("utf-8").split("\0")
        return {Path(file_) for file_ in all_files}

    def _find_submodules(self) -> set[Path]:
        command = [
            str(self.EXE),
            "config",
            "-z",
            "--file",
            ".gitmodules",
            "--get-regexp",
            r"\.path$",
        ]
        result = execute_command(command, _LOGGER, cwd=self.root)
        # The final element may be an empty string. Filter it.
        submodule_entries = [
            entry
            for entry in result.stdout.decode("utf-8").split("\0")
            if entry
        ]
        # Each entry looks a little like 'submodule.submodule.path\nmy_path'.
        return {Path(entry.splitlines()[1]) for entry in submodule_entries}

    def is_ignored(self, path: Path) -> bool:
        path = relative_from_root(path, self.root)
        return path in self._all_ignored_files

    def is_submodule(self, path: StrPath) -> bool:
        return any(
            relative_from_root(Path(path), self.root).resolve()
            == submodule_path.resolve()
            for submodule_path in self._submodules
        )

    @classmethod
    def in_repo(cls, directory: StrPath) -> bool:
        if not Path(directory).is_dir():
            raise NotADirectoryError()

        if _find_ancestor(directory, ".git", is_directory=False):
            command = [str(cls.EXE), "rev-parse", "--is-inside-work-tree"]
            result = execute_command(command, _LOGGER, cwd=directory)

            return not result.returncode
        return False

    @classmethod
    def find_root(cls, cwd: StrPath | None = None) -> Path | None:
        if cwd is None:
            cwd = Path.cwd()

        if not Path(cwd).is_dir():
            raise NotADirectoryError()
        if not _find_ancestor(cwd, ".git", is_directory=False):
            return None

        command = [str(cls.EXE), "rev-parse", "--show-toplevel"]
        result = execute_command(command, _LOGGER, cwd=cwd)

        if not result.returncode:
            path = result.stdout.decode("utf-8")[:-1]
            return Path(os.path.relpath(path, cwd))

        return None


class VCSStrategyHg(VCSStrategy):
    """Strategy that is used for Mercurial."""

    EXE = HG_EXE

    def __init__(self, root: StrPath):
        super().__init__(root)
        if not self.EXE:
            raise FileNotFoundError("Could not find binary for Mercurial")
        self._all_ignored_files = self._find_all_ignored_files()

    def _find_all_ignored_files(self) -> set[Path]:
        """Return a set of all files ignored by mercurial. If a whole directory
        is ignored, don't return all files inside of it.
        """
        command = [
            str(self.EXE),
            "status",
            "--ignored",
            # terse is marked 'experimental' in the hg help but is documented
            # in the man page. It collapses the output of a dir containing only
            # ignored files to the ignored name like the git command does.
            # TODO: Re-enable this flag in the future.
            # "--terse=i",
            "--no-status",
            "--print0",
        ]
        result = execute_command(command, _LOGGER, cwd=self.root)
        all_files = result.stdout.decode("utf-8").split("\0")
        return {Path(file_) for file_ in all_files}

    def is_ignored(self, path: Path) -> bool:
        path = relative_from_root(path, self.root)
        return path in self._all_ignored_files

    def is_submodule(self, path: StrPath) -> bool:
        # TODO: Implement me.
        return False

    @classmethod
    def in_repo(cls, directory: StrPath) -> bool:
        if not Path(directory).is_dir():
            raise NotADirectoryError()

        if _find_ancestor(directory, ".hg"):
            command = [str(cls.EXE), "root"]
            result = execute_command(command, _LOGGER, cwd=directory)

            return not result.returncode
        return False

    @classmethod
    def find_root(cls, cwd: StrPath | None = None) -> Path | None:
        if cwd is None:
            cwd = Path.cwd()

        if not Path(cwd).is_dir():
            raise NotADirectoryError()
        if not _find_ancestor(cwd, ".hg"):
            return None

        command = [str(cls.EXE), "root"]
        result = execute_command(command, _LOGGER, cwd=cwd)

        if not result.returncode:
            path = result.stdout.decode("utf-8")[:-1]
            return Path(os.path.relpath(path, cwd))

        return None


class VCSStrategyJujutsu(VCSStrategy):
    """Strategy that is used for Jujutsu."""

    EXE = JUJUTSU_EXE

    def __init__(self, root: StrPath):
        super().__init__(root)
        if not self.EXE:
            raise FileNotFoundError("Could not find binary for Jujutsu")
        self._all_tracked_files = self._find_all_tracked_files()

    def _find_all_tracked_files(self) -> set[Path]:
        """
        Return a set of all files tracked in the current jj revision
        """
        version = self._version()
        # TODO: Remove the version check once most distributions ship jj 0.19.0
        # or higher.
        if version is None or version >= (0, 19, 0):
            command = [str(self.EXE), "file", "list"]
        else:
            command = [str(self.EXE), "files"]
        result = execute_command(command, _LOGGER, cwd=self.root)
        all_files = result.stdout.decode("utf-8").split("\n")
        return {Path(file_) for file_ in all_files if file_}

    def _version(self) -> tuple[int, int, int] | None:
        """
        Returns the (major, minor, patch) version of the jujutsu executable,
        or None if the version components cannot be determined.
        """
        result = execute_command(
            [str(self.EXE), "--version"], _LOGGER, cwd=self.root
        )
        lines = result.stdout.decode("utf-8").split("\n")
        # Output has the form `jj major.minor.patch[-hash]\n`.
        try:
            line = lines[0]
            version = line.split(" ")[-1]
            without_hash = version.split("-")[0]
            components = without_hash.split(".")
            return (int(components[0]), int(components[1]), int(components[2]))
        except (IndexError, ValueError) as e:
            _LOGGER.debug("unable to parse jj version: %s", e)
            return None

    def is_ignored(self, path: Path) -> bool:
        path = relative_from_root(path, self.root)

        for tracked in self._all_tracked_files:
            if tracked.parts[: len(path.parts)] == path.parts:
                # We can't check only if the path is in our tracked files as we
                # must support directories as well as files
                #
                # We'll consider a directory "tracked" if there are any tracked
                # files inside it
                return False

        return True

    def is_submodule(self, path: StrPath) -> bool:
        return False

    @classmethod
    def in_repo(cls, directory: StrPath) -> bool:
        if not Path(directory).is_dir():
            raise NotADirectoryError()

        if _find_ancestor(directory, ".jj"):
            command = [str(cls.EXE), "root"]
            result = execute_command(command, _LOGGER, cwd=directory)

            return not result.returncode
        return False

    @classmethod
    def find_root(cls, cwd: StrPath | None = None) -> Path | None:
        if cwd is None:
            cwd = Path.cwd()

        if not Path(cwd).is_dir():
            raise NotADirectoryError()
        if not _find_ancestor(cwd, ".jj"):
            return None

        command = [str(cls.EXE), "root"]
        result = execute_command(command, _LOGGER, cwd=cwd)

        if not result.returncode:
            path = result.stdout.decode("utf-8")[:-1]
            return Path(os.path.relpath(path, cwd))

        return None


class VCSStrategyPijul(VCSStrategy):
    """Strategy that is used for Pijul."""

    EXE = PIJUL_EXE

    def __init__(self, root: StrPath):
        super().__init__(root)
        if not self.EXE:
            raise FileNotFoundError("Could not find binary for Pijul")
        self._all_tracked_files = self._find_all_tracked_files()

    def _find_all_tracked_files(self) -> set[Path]:
        """Return a set of all files tracked by pijul."""
        command = [str(self.EXE), "list"]
        result = execute_command(command, _LOGGER, cwd=self.root)
        all_files = result.stdout.decode("utf-8").splitlines()
        return {Path(file_) for file_ in all_files}

    def is_ignored(self, path: Path) -> bool:
        path = relative_from_root(path, self.root)
        return path not in self._all_tracked_files

    def is_submodule(self, path: StrPath) -> bool:
        # not supported in pijul yet
        return False

    @classmethod
    def in_repo(cls, directory: StrPath) -> bool:
        if not Path(directory).is_dir():
            raise NotADirectoryError()

        if _find_ancestor(directory, ".pijul"):
            command = [str(cls.EXE), "diff", "--short"]
            result = execute_command(command, _LOGGER, cwd=directory)

            return not result.returncode
        return False

    @classmethod
    def find_root(cls, cwd: StrPath | None = None) -> Path | None:
        if cwd is None:
            cwd = Path.cwd()

        # TODO this duplicates pijul's logic.
        # Maybe it should be replaced by calling pijul,
        # but there is no matching subcommand yet.
        path = Path(cwd).resolve()

        if not path.is_dir():
            raise NotADirectoryError()

        dot_pijul = _find_ancestor(path, ".pijul")
        if dot_pijul is not None:
            return dot_pijul.parent
        return None


def all_vcs_strategies() -> Generator[type[VCSStrategy]]:
    """Yield all VCSStrategy classes that aren't the abstract base class."""
    for value in globals().values():
        if (
            isclass(value)
            and issubclass(value, VCSStrategy)
            and value is not VCSStrategy
        ):
            yield value


def find_root(cwd: StrPath | None = None) -> Path | None:
    """Try to find the root of the project from *cwd*. If none is found,
    return None.

    Raises:
        NotADirectoryError: if directory is not a directory.
    """
    for strategy in all_vcs_strategies():
        if strategy.EXE:
            root = strategy.find_root(cwd=cwd)
            if root:
                return root
    return None



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2023 Carmen Bianca BAKKER <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2023 Matthias Riße
# SPDX-FileCopyrightText: 2023 DB Systel GmbH
# SPDX-FileCopyrightText: 2024 Kerry McAdams <github@klmcadams>
# SPDX-FileCopyrightText: 2024 Linnea Gräf
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Module that contains the central Project class."""

import errno
import glob
import logging
import os
import warnings
from collections import defaultdict
from collections.abc import Collection, Iterator
from pathlib import Path
from typing import NamedTuple

import attrs

from ._licenses import EXCEPTION_MAP, LICENSE_MAP
from ._util import _determine_license_path, relative_from_root
from .copyright import ReuseInfo, SourceType
from .covered_files import iter_files
from .exceptions import (
    GlobalLicensingConflictError,
    SpdxIdentifierNotFoundError,
)
from .extract import _LICENSEREF_PATTERN, CHUNK_SIZE, reuse_info_of_file
from .global_licensing import (
    GlobalLicensing,
    NestedReuseTOML,
    PrecedenceType,
    ReuseDep5,
    ReuseTOML,
)
from .i18n import _
from .types import StrPath
from .vcs import VCSStrategy, VCSStrategyNone, all_vcs_strategies

_LOGGER = logging.getLogger(__name__)


class GlobalLicensingFound(NamedTuple):
    path: Path
    cls: type[GlobalLicensing]


# TODO: The information (root, include_submodules, include_meson_subprojects,
# vcs_strategy) is passed to SO MANY PLACES. Maybe Project should be simplified
# to contain exclusively those values, or maybe these values should be extracted
# out of Project to simplify passing this information around.
@attrs.define
class Project:
    """Simple object that holds the project's root, which is necessary for many
    interactions.
    """

    root: Path = attrs.field(converter=Path)
    include_submodules: bool = False
    include_meson_subprojects: bool = False
    vcs_strategy: VCSStrategy = attrs.field()
    global_licensing: GlobalLicensing | None = None

    # TODO: I want to get rid of these, or somehow refactor this mess.
    license_map: dict[str, dict] = attrs.field()
    licenses: dict[str, Path] = attrs.field(factory=dict)

    licenses_without_extension: dict[str, Path] = attrs.field(
        init=False, factory=dict
    )

    @vcs_strategy.default
    def _default_vcs_strategy(self) -> VCSStrategy:
        return VCSStrategyNone(self.root)

    @license_map.default
    def _default_license_map(self) -> dict[str, dict]:
        license_map = LICENSE_MAP.copy()
        license_map.update(EXCEPTION_MAP)
        return license_map

    @classmethod
    def from_directory(
        cls,
        root: StrPath,
        include_submodules: bool = False,
        include_meson_subprojects: bool = False,
    ) -> "Project":
        """A factory method that reads various files in the *root* directory to
        correctly build the :class:`Project` object.

        Args:
            root: The root of the project.
            include_submodules: Whether to also lint VCS submodules.
            include_meson_subprojects: Whether to also lint Meson subprojects.

        Raises:
            FileNotFoundError: if root does not exist.
            NotADirectoryError: if root is not a directory.
            UnicodeDecodeError: if the global licensing config file could not be
                decoded.
            GlobalLicensingParseError: if the global licensing config file could
                not be parsed.
            GlobalLicensingConflictError: if more than one global licensing
                config file is present.
        """
        root = Path(root)
        if not root.exists():
            raise FileNotFoundError(
                errno.ENOENT,
                os.strerror(errno.ENOENT),
                str(root),
            )
        if not root.is_dir():
            raise NotADirectoryError(
                errno.ENOTDIR,
                os.strerror(errno.ENOTDIR),
                str(root),
            )

        vcs_strategy = cls._detect_vcs_strategy(root)

        global_licensing: GlobalLicensing | None = None
        found = cls.find_global_licensing(
            root,
            include_submodules=include_submodules,
            include_meson_subprojects=include_meson_subprojects,
            vcs_strategy=vcs_strategy,
        )
        if found:
            global_licensing = cls._global_licensing_from_found(
                found, str(root)
            )

        project = cls(
            root,
            vcs_strategy=vcs_strategy,
            global_licensing=global_licensing,
            include_submodules=include_submodules,
            include_meson_subprojects=include_meson_subprojects,
        )

        # TODO: Because the `_find_licenses()` method is so broad and depends on
        # some object attributes, we set the attribute after creating the
        # object. Ideally we do this before creating the object, but that would
        # require refactoring the method.
        project.licenses = project._find_licenses()

        return project

    def all_files(self, directory: StrPath | None = None) -> Iterator[Path]:
        """Yield all files in *directory* and its subdirectories.

        The files that are not yielded are those explicitly ignored by the REUSE
        Specification. That means:

        - LICENSE/COPYING files.
        - VCS directories.
        - .license files.
        - .spdx files.
        - Files ignored by VCS.
        - Symlinks.
        - Submodules (depending on the value of :attr:`include_submodules`).
        - Meson subprojects (depending on the value of
              :attr:`include_meson_subprojects`).
        - 0-sized files.

        Args:
            directory: The directory in which to search.
        """
        if directory is None:
            directory = self.root
        return iter_files(
            directory,
            include_submodules=self.include_submodules,
            include_meson_subprojects=self.include_meson_subprojects,
            vcs_strategy=self.vcs_strategy,
        )

    def subset_files(
        self, files: Collection[StrPath], directory: StrPath | None = None
    ) -> Iterator[Path]:
        """Like :meth:`all_files`, but all files that are not in *files* are
        filtered out.

        Args:
            files: A collection of paths relative to the current working
                directory. Any files that are not in this collection are not
                yielded.
            directory: The directory in which to search.
        """
        if directory is None:
            directory = self.root
        return iter_files(
            directory=directory,
            subset_files=files,
            include_submodules=self.include_submodules,
            include_meson_subprojects=self.include_meson_subprojects,
            vcs_strategy=self.vcs_strategy,
        )

    def reuse_info_of(self, path: StrPath) -> list[ReuseInfo]:
        """Return REUSE info of *path*.

        This function will return any REUSE information that it can find: from
        within the file, the .license file, from REUSE.toml, and/or from the
        .reuse/dep5 file.

        The presence of a .license file always means that the file itself will
        not be parsed for REUSE information.

        When information is found from multiple sources, and if the precedence
        for that file in REUSE.toml is 'aggregate' (or if .reuse/dep5 is used),
        then two (or more) :class:`ReuseInfo` objects are returned in list set,
        each with respective discovered REUSE information and information about
        the source.

        Alternatively, if the precedence is set to 'closest' or 'toml', or if
        information was found in only one source, then a list of one item is
        returned.

        The exact precedence handling is detailed in the specification.

        An empty list is returned if no information was found whatsoever.
        """
        # pylint: disable=too-many-branches
        original_path = Path(path)
        path = _determine_license_path(path)

        # This means that only one 'source' of licensing/copyright information
        # is captured in ReuseInfo
        global_results: defaultdict[PrecedenceType, list[ReuseInfo]] = (
            defaultdict(list)
        )
        file_result = ReuseInfo()
        result: list[ReuseInfo] = []

        # Search the global licensing file for REUSE information.
        if self.global_licensing:
            relpath = self.relative_from_root(original_path)
            global_results = defaultdict(
                list, self.global_licensing.reuse_info_of(relpath)
            )
            for info_list in global_results.values():
                for global_result in info_list:
                    if global_result.contains_copyright_or_licensing():
                        _LOGGER.info(
                            _("'{path}' covered by {global_path}").format(
                                path=path, global_path=global_result.source_path
                            )
                        )

        if PrecedenceType.OVERRIDE in global_results:
            _LOGGER.info(
                _(
                    "'{path}' is covered exclusively by REUSE.toml. Not reading"
                    " the file contents."
                ).format(path=path)
            )
        else:
            with path.open("rb", buffering=CHUNK_SIZE) as fp:
                file_result = reuse_info_of_file(fp)
            if file_result.contains_info():
                source_type = SourceType.FILE_HEADER
                if path.suffix == ".license":
                    source_type = SourceType.DOT_LICENSE
                file_result = file_result.copy(
                    path=relative_from_root(
                        original_path, self.root
                    ).as_posix(),
                    source_path=relative_from_root(path, self.root).as_posix(),
                    source_type=source_type,
                )

        result.extend(global_results[PrecedenceType.OVERRIDE])
        result.extend(global_results[PrecedenceType.AGGREGATE])
        if file_result.contains_info():
            result.append(file_result)
        if not file_result.contains_copyright_or_licensing():
            result.extend(global_results[PrecedenceType.CLOSEST])
        # Special case: If a file contains only copyright, apply the
        # REUSE.toml's licensing if it exists, and vice versa.
        elif file_result.contains_copyright_xor_licensing():
            if global_results[PrecedenceType.CLOSEST]:
                # There should only by a single CLOSEST result in the list.
                closest = global_results[PrecedenceType.CLOSEST][0]
                if file_result.copyright_notices:
                    result.append(
                        closest.copy(
                            copyright_notices=set(),
                        )
                    )
                elif file_result.spdx_expressions:
                    result.append(
                        closest.copy(
                            spdx_expressions=set(),
                        )
                    )
        return result

    def relative_from_root(self, path: Path) -> Path:
        """If the project root is /tmp/project, and *path* is
        /tmp/project/src/file, then return src/file.
        """
        return relative_from_root(path, self.root)

    @classmethod
    def find_global_licensing(
        cls,
        root: Path,
        include_submodules: bool = False,
        include_meson_subprojects: bool = False,
        vcs_strategy: VCSStrategy | None = None,
    ) -> list[GlobalLicensingFound]:
        """Find the path and corresponding class of a project directory's
        :class:`GlobalLicensing`.

        Raises:
            GlobalLicensingConflictError: if more than one global licensing
                config file is present.
        """
        candidates: list[GlobalLicensingFound] = []
        dep5_path = root / ".reuse/dep5"
        if (dep5_path).exists():
            # Sneaky workaround to not print this warning.
            if not os.environ.get("_SUPPRESS_DEP5_WARNING"):
                warnings.warn(
                    _(
                        "'.reuse/dep5' is deprecated. You are recommended to"
                        " instead use REUSE.toml. Use `reuse convert-dep5` to"
                        " convert."
                    ),
                    PendingDeprecationWarning,
                )
            candidates = [GlobalLicensingFound(dep5_path, ReuseDep5)]

        reuse_toml_candidates = [
            GlobalLicensingFound(path, ReuseTOML)
            for path in NestedReuseTOML.find_reuse_tomls(
                root,
                include_submodules=include_submodules,
                include_meson_subprojects=include_meson_subprojects,
                vcs_strategy=vcs_strategy,
            )
        ]
        if reuse_toml_candidates:
            if candidates:
                raise GlobalLicensingConflictError(
                    _(
                        "Found both '{new_path}' and '{old_path}'. You"
                        " cannot keep both files simultaneously; they are"
                        " not intercompatible."
                    ).format(
                        new_path=reuse_toml_candidates[0].path,
                        old_path=dep5_path,
                    )
                )
            candidates = reuse_toml_candidates

        return candidates

    @classmethod
    def _global_licensing_from_found(
        cls, found: list[GlobalLicensingFound], root: StrPath
    ) -> GlobalLicensing:
        if len(found) == 1 and found[0].cls == ReuseDep5:
            return ReuseDep5.from_file(found[0].path)
        # This is an impossible scenario at time of writing.
        if not all(item.cls == ReuseTOML for item in found):
            raise NotImplementedError()
        tomls = [ReuseTOML.from_file(item.path) for item in found]
        return NestedReuseTOML(reuse_tomls=tomls, source=str(root))

    def _identifier_of_license(self, path: Path) -> str:
        """Figure out the SPDX License identifier of a license given its path.
        The name of the path (minus its extension) should be a valid SPDX
        License Identifier.
        """
        if not path.suffix:
            raise SpdxIdentifierNotFoundError(f"{path} has no file extension")
        if path.stem in self.license_map:
            return path.stem
        if _LICENSEREF_PATTERN.match(path.stem):
            return path.stem

        raise SpdxIdentifierNotFoundError(
            f"Could not find SPDX License Identifier for {path}"
        )

    def _find_licenses(self) -> dict[str, Path]:
        """Return a dictionary of all licenses in the project, with their SPDX
        identifiers as names and paths as values.
        """
        # TODO: This method does more than one thing. We ought to simplify it.
        license_files: dict[str, Path] = {}

        directory = str(self.root / "LICENSES/**")
        for path_str in glob.iglob(directory, recursive=True):
            path = Path(path_str)
            # For some reason, LICENSES/** is resolved even though it
            # doesn't exist. I have no idea why. Deal with that here.
            if not Path(path).exists() or Path(path).is_dir():
                continue
            if Path(path).suffix == ".license":
                continue

            # path = self.relative_from_root(path)

            try:
                identifier = self._identifier_of_license(path)
            except SpdxIdentifierNotFoundError:
                if path.name in self.license_map:
                    _LOGGER.info(
                        _("{path} does not have a file extension").format(
                            path=path
                        )
                    )
                    identifier = path.name
                    self.licenses_without_extension[identifier] = path
                else:
                    identifier = path.stem
                    _LOGGER.warning(
                        _(
                            "Could not resolve SPDX License Identifier of"
                            " {path}, resolving to {identifier}. Make sure the"
                            " license is in the license list found at"
                            " <https://spdx.org/licenses/> or that it starts"
                            " with 'LicenseRef-', and that it has a file"
                            " extension."
                        ).format(path=path, identifier=identifier)
                    )

            if identifier in license_files:
                _LOGGER.critical(
                    _(
                        "{identifier} is the SPDX License Identifier of both"
                        " {path} and {other_path}"
                    ).format(
                        identifier=identifier,
                        path=path,
                        other_path=license_files[identifier],
                    )
                )
                raise RuntimeError("Multiple licenses resolve to {identifier}")
            # Add the identifiers
            license_files[identifier] = path
            if (
                _LICENSEREF_PATTERN.match(identifier)
                and "Unknown" not in identifier
            ):
                self.license_map[identifier] = {
                    "reference": str(path),
                    "isDeprecatedLicenseId": False,
                    "detailsUrl": None,
                    "referenceNumber": None,
                    "name": identifier,
                    "licenseId": identifier,
                    "seeAlso": [],
                    "isOsiApproved": None,
                }

        return license_files

    @classmethod
    def _detect_vcs_strategy(cls, root: StrPath) -> VCSStrategy:
        """For each supported VCS, check if the software is available and if the
        directory is a repository. If not, return :class:`VCSStrategyNone`.
        """
        for strategy in all_vcs_strategies():
            if strategy.EXE and strategy.in_repo(root):
                return strategy(root)

        _LOGGER.info(
            _(
                "project '{}' is not a VCS repository or required VCS"
                " software is not installed"
            ).format(root)
        )
        return VCSStrategyNone(root)



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Entry module for reuse."""

if __name__ == "__main__":
    from .cli.main import main

    # pylint: disable=no-value-for-parameter
    main()



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""All exceptions owned by :mod:`reuse`. These exceptions all inherit
:class:`ReuseError`.
"""

from typing import Any


class ReuseError(Exception):
    """Base exception."""


class SpdxIdentifierNotFoundError(ReuseError):
    """Could not find SPDX identifier for license file."""


class GlobalLicensingParseError(ReuseError):
    """An exception representing any kind of error that occurs when trying to
    parse a :class:`reuse.global_licensing.GlobalLicensing` file.
    """

    def __init__(self, *args: Any, source: str | None = None):
        super().__init__(*args)
        self.source = source


class GlobalLicensingParseTypeError(GlobalLicensingParseError, TypeError):
    """An exception representing a type error while trying to parse a
    :class:`reuse.global_licensing.GlobalLicensing` file.
    """


class GlobalLicensingParseValueError(GlobalLicensingParseError, ValueError):
    """An exception representing a value error while trying to parse a
    :class:`reuse.global_licensing.GlobalLicensing` file.
    """


class GlobalLicensingConflictError(ReuseError):
    """There are two global licensing files in the project that are not
    compatible.
    """


class MissingReuseInfoError(ReuseError):
    """Some REUSE information is missing from the result."""


class CommentError(ReuseError):
    """An error occurred during an interaction with a comment."""


class CommentCreateError(CommentError):
    """An error occurred during the creation of a comment."""


class CommentParseError(CommentError):
    """An error occurred during the parsing of a comment."""


class CopyrightNoticeParseError(ReuseError):
    """Error for parsing a :class:`reuse.copyright.CopyrightNotice` from a
    string.
    """


class YearRangeParseError(CopyrightNoticeParseError):
    """Error for parsing a :class:`reuse.copyright.YearRange` from a string.
    Because a year range is typically a constituent part of a copyright notice,
    this error is a subclass of :class:`CopyrightNoticeParseError`.
    """


class NoEncodingModuleError(ReuseError):
    """No module that can detect the encoding of a file is importable."""



# SPDX-FileCopyrightText: 2019 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2019 Kirill Elagin <kirelagin@gmail.com>
# SPDX-FileCopyrightText: 2019 Stefan Bakker <s.bakker777@gmail.com>
# SPDX-FileCopyrightText: 2020 Dmitry Bogatov
# SPDX-FileCopyrightText: 2021 Alliander N.V. <https://alliander.com>
# SPDX-FileCopyrightText: 2021 Alvar Penning
# SPDX-FileCopyrightText: 2021 Robin Vobruba <hoijui.quaero@gmail.com>
# SPDX-FileCopyrightText: 2022 Carmen Bianca Bakker <carmenbianca@fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2022 Yaman Qalieh
# SPDX-FileCopyrightText: 2024 Rivos Inc.
# SPDX-FileCopyrightText: © 2020 Liferay, Inc. <https://liferay.com>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""Click code for annotate subcommand."""

import datetime
import logging
import os
import sys
from collections.abc import Collection, Iterable, Sequence
from pathlib import Path
from typing import Any, cast

import click
from jinja2 import Environment, FileSystemLoader, Template
from jinja2.exceptions import TemplateNotFound

from .._annotate import add_header_to_file
from .._util import _determine_license_path, _determine_license_suffix_path
from ..comment import (
    NAME_STYLE_MAP,
    CommentStyle,
    get_comment_style,
    has_style,
    is_uncommentable,
)
from ..copyright import (
    CopyrightNotice,
    CopyrightPrefix,
    ReuseInfo,
    SpdxExpression,
    YearRange,
    validate_four_digits,
)
from ..exceptions import YearRangeParseError
from ..extract import HEURISTICS_CHUNK_SIZE, detect_encoding, detect_newline
from ..i18n import _
from ..project import Project
from .common import ClickObj, MutexOption, spdx_identifier
from .main import main

_LOGGER = logging.getLogger(__name__)


def test_mandatory_option_required(
    copyright_: Any,
    license_: Any,
    contributor: Any,
) -> None:
    """Raise a parser error if one of the mandatory options is not provided."""
    if not any((copyright_, license_, contributor)):
        raise click.UsageError(
            _(
                "Option '--copyright', '--license', or '--contributor' is"
                " required."
            )
        )


def all_paths(
    paths: Collection[Path],
    recursive: bool,
    project: Project,
) -> list[Path]:
    """Return a set of all provided paths, converted into .license paths if they
    exist. If *recursive* is enabled, all files belonging to *project* that are
    recursive children of *paths* are also added.

    Directories are filtered out.
    """
    if recursive:
        result: set[Path] = set()
        all_files = [path.resolve() for path in project.all_files()]
        for path in paths:
            if path.is_file():
                result.add(path)
            else:
                result |= {
                    child
                    for child in all_files
                    if path.resolve() in child.parents
                }
    else:
        result = set(paths)
    return [_determine_license_path(path) for path in result if path.is_file()]


def verify_paths_comment_style(
    style: Any,
    fallback_dot_license: Any,
    skip_unrecognised: Any,
    force_dot_license: Any,
    paths: Iterable[Path],
) -> None:
    """Exit if --style, --force-dot-license, --fallback-dot-license,
    or --skip-unrecognised is not enabled and one of the paths has an
    unrecognised style.
    """
    if (
        not style
        and not fallback_dot_license
        and not skip_unrecognised
        and not force_dot_license
    ):
        unrecognised_files: set[Path] = set()

        for path in paths:
            if not has_style(path):
                unrecognised_files.add(path)

        if unrecognised_files:
            raise click.UsageError(
                "{}\n\n{}".format(
                    _(
                        "The following files do not have a recognised file"
                        " extension. Please use '--style',"
                        " '--force-dot-license', '--fallback-dot-license', or"
                        " '--skip-unrecognised':"
                    ),
                    "\n".join(str(path) for path in unrecognised_files),
                )
            )


def verify_paths_line_handling(
    single_line: bool,
    multi_line: bool,
    forced_style: str | None,
    paths: Iterable[Path],
) -> None:
    """This function aborts the parser when --single-line or --multi-line is
    used, but the file type does not support that type of comment style.
    """
    for path in paths:
        style: type[CommentStyle] | None = None
        if forced_style is not None:
            style = NAME_STYLE_MAP.get(forced_style)
        if style is None:
            style = get_comment_style(path)
        # This shouldn't happen because of prior tests, so let's not bother with
        # this case.
        if style is None:
            continue
        # TODO: list all non-functional paths
        if single_line and not style.can_handle_single():
            raise click.UsageError(
                _(
                    "'{path}' does not support single-line comments, please"
                    " do not use '--single-line'."
                ).format(path=path.as_posix())
            )
        if multi_line and not style.can_handle_multi():
            raise click.UsageError(
                _(
                    "'{path}' does not support multi-line comments, please"
                    " do not use '--multi-line'."
                ).format(path=path.as_posix())
            )


def find_template(project: Project, name: str) -> Template:
    """Find a template given a name.

    Raises:
        TemplateNotFound: if template could not be found.
    """
    template_dir = project.root / ".reuse/templates"
    env = Environment(
        loader=FileSystemLoader(str(template_dir)), trim_blocks=True
    )

    names = [name]
    if not name.endswith(".jinja2"):
        names.append(f"{name}.jinja2")
    if not name.endswith(".commented.jinja2"):
        names.append(f"{name}.commented.jinja2")

    for item in names:
        try:
            return env.get_template(item)
        except TemplateNotFound:
            pass
    raise TemplateNotFound(name)


def get_template(
    template_str: str | None, project: Project
) -> tuple[Template | None, bool]:
    """If a template is specified on the CLI, find and return it, including
    whether it is a 'commented' template.

    If no template is specified, just return None.
    """
    template: Template | None = None
    commented = False
    if template_str:
        try:
            template = cast(Template, find_template(project, template_str))
        except TemplateNotFound as error:
            raise click.UsageError(
                _("Template '{template}' could not be found.").format(
                    template=template_str
                )
            ) from error

        if ".commented" in Path(cast(str, template.name)).suffixes:
            commented = True
    return template, commented


def get_years(year: str | None, exclude_year: bool) -> tuple[YearRange, ...]:
    """Get the year. Normally it is today's year. If --year is specified,
    get a full tuple of ranges from that one.

    If --exclude-year is specified, return an empty tuple.
    """
    result: tuple[YearRange, ...] = tuple()
    if not exclude_year:
        if year:
            try:
                result = YearRange.tuple_from_string(year)
            except YearRangeParseError as error:
                raise click.UsageError(
                    _("'{year}' is not a valid year range.").format(year=year)
                ) from error
        else:
            try:
                current_year = str(datetime.date.today().year)
                result = (YearRange(validate_four_digits(current_year)),)
            except ValueError as error:
                raise click.UsageError(
                    _(
                        "Your operating system's year is set to '{year}'. This"
                        " is not four digits, and not supported."
                    ).format(year=current_year)
                ) from error
    return result


def get_reuse_info(
    copyrights: Collection[str],
    licenses: Collection[SpdxExpression],
    contributors: Collection[str],
    copyright_prefix: str | None,
    years: tuple[YearRange, ...],
) -> ReuseInfo:
    """Create a ReuseInfo object from --license, --copyright, and
    --contributor.
    """
    prefix = (
        CopyrightPrefix[CopyrightPrefix.uppercase_name(copyright_prefix)]
        if copyright_prefix is not None
        else CopyrightPrefix.SPDX
    )
    copyright_notices = {
        CopyrightNotice(item, years=years, prefix=prefix) for item in copyrights
    }

    return ReuseInfo(
        spdx_expressions=set(licenses),
        copyright_notices=copyright_notices,
        contributor_lines=set(contributors),
    )


_YEAR_MUTEX = ["years", "exclude_year"]
_LINE_MUTEX = ["single_line", "multi_line"]
_STYLE_MUTEX = [
    "force_dot_license",
    "fallback_dot_license",
    "skip_unrecognised",
]

_HELP = (
    _("Add copyright and licensing into the headers of files.")
    + "\n\n"
    + _(
        "By using --copyright and --license, you can specify which"
        " copyright holders and licenses to add to the headers of the"
        " given files."
    )
    + "\n\n"
    + _(
        "By using --contributor, you can specify people or entity that"
        " contributed but are not copyright holder of the given"
        " files."
    )
)


@main.command(name="annotate", help=_HELP)
@click.option(
    "--copyright",
    "-c",
    "copyrights",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("COPYRIGHT"),
    type=str,
    multiple=True,
    help=_("Copyright holder, repeatable."),
)
@click.option(
    "--license",
    "-l",
    "licenses",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("SPDX_IDENTIFIER"),
    type=spdx_identifier,
    multiple=True,
    help=_("SPDX License Identifier, repeatable."),
)
@click.option(
    "--contributor",
    "contributors",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("CONTRIBUTOR"),
    type=str,
    multiple=True,
    help=_("File contributor, repeatable."),
)
@click.option(
    "--year",
    "-y",
    "years",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("YEAR"),
    cls=MutexOption,
    mutually_exclusive=_YEAR_MUTEX,
    type=str,
    help=_(
        "Year of copyright notice. You may define multiple years or a range"
        " of years."
    ),
)
@click.option(
    "--style",
    "-s",
    cls=MutexOption,
    mutually_exclusive=["skip_unrecognised"],
    type=click.Choice(list(NAME_STYLE_MAP)),
    help=_("Comment style to use."),
)
@click.option(
    "--copyright-prefix",
    type=click.Choice(
        [
            CopyrightPrefix.lowercase_name(prefix.name)
            for prefix in CopyrightPrefix
        ]
    ),
    help=_("Copyright prefix to use."),
)
@click.option(
    "--copyright-style",
    "copyright_prefix",
    hidden=True,
)
@click.option(
    "--template",
    "-t",
    "template_str",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("TEMPLATE"),
    type=str,
    help=_("Name of template to use."),
)
@click.option(
    "--exclude-year",
    cls=MutexOption,
    mutually_exclusive=_YEAR_MUTEX,
    is_flag=True,
    help=_("Do not include year in copyright notice."),
)
@click.option(
    "--merge-copyrights",
    is_flag=True,
    help=_(
        "Merge copyright notices if they are identical except for their years."
    ),
)
@click.option(
    "--single-line",
    cls=MutexOption,
    mutually_exclusive=_LINE_MUTEX,
    is_flag=True,
    help=_("Force single-line comment style."),
)
@click.option(
    "--multi-line",
    cls=MutexOption,
    mutually_exclusive=_LINE_MUTEX,
    is_flag=True,
    help=_("Force multi-line comment style."),
)
@click.option(
    "--recursive",
    "-r",
    is_flag=True,
    help=_("Add headers to all files under specified directories recursively."),
)
@click.option(
    "--no-replace",
    is_flag=True,
    help=_("Do not replace the first header in the file; just add a new one."),
)
@click.option(
    "--force-dot-license",
    cls=MutexOption,
    mutually_exclusive=_STYLE_MUTEX,
    is_flag=True,
    help=_("Always write a .license file instead of a header inside the file."),
)
@click.option(
    "--fallback-dot-license",
    cls=MutexOption,
    mutually_exclusive=_STYLE_MUTEX,
    is_flag=True,
    help=_("Write a .license file to files with unrecognised comment styles."),
)
@click.option(
    "--skip-unrecognised",
    cls=MutexOption,
    mutually_exclusive=_STYLE_MUTEX,
    is_flag=True,
    help=_("Skip files with unrecognised comment styles."),
)
@click.option(
    "--skip-unrecognized",
    "skip_unrecognised",
    is_flag=True,
    hidden=True,
)
@click.option(
    "--skip-existing",
    is_flag=True,
    help=_("Skip files that already contain REUSE information."),
)
@click.argument(
    "paths",
    # TRANSLATORS: You may translate this. Please preserve capital letters.
    metavar=_("PATH"),
    type=click.Path(exists=True, writable=True, path_type=Path),
    nargs=-1,
)
@click.pass_obj
def annotate(
    obj: ClickObj,
    copyrights: Sequence[str],
    licenses: Sequence[SpdxExpression],
    contributors: Sequence[str],
    years: str | None,
    style: str | None,
    copyright_prefix: str | None,
    template_str: str | None,
    exclude_year: bool,
    merge_copyrights: bool,
    single_line: bool,
    multi_line: bool,
    recursive: bool,
    no_replace: bool,
    force_dot_license: bool,
    fallback_dot_license: bool,
    skip_unrecognised: bool,
    skip_existing: bool,
    paths: Sequence[Path],
) -> None:
    # pylint: disable=too-many-arguments,too-many-locals,missing-function-docstring
    project = obj.project

    test_mandatory_option_required(copyrights, licenses, contributors)
    paths = all_paths(paths, recursive, project)
    verify_paths_comment_style(
        style, fallback_dot_license, skip_unrecognised, force_dot_license, paths
    )
    # Verify line handling and comment styles before proceeding.
    verify_paths_line_handling(single_line, multi_line, style, paths)
    template, commented = get_template(template_str, project)
    years_tuple = get_years(years, exclude_year)
    reuse_info = get_reuse_info(
        copyrights, licenses, contributors, copyright_prefix, years_tuple
    )

    result = 0
    for path in paths:
        with path.open("rb") as fp:
            chunk = fp.read(HEURISTICS_CHUNK_SIZE)
        encoding = detect_encoding(chunk)
        newline = (
            detect_newline(chunk, encoding=encoding)
            if encoding is not None
            else os.linesep
        )
        if encoding is None or is_uncommentable(path) or force_dot_license:
            new_path = _determine_license_suffix_path(path)
            if encoding is None:
                _LOGGER.info(
                    _(
                        "'{path}' is a binary, therefore using '{new_path}'"
                        " for the header"
                    ).format(path=path, new_path=new_path)
                )
                encoding = "utf_8"
            path = Path(new_path)
            path.touch()
        result += add_header_to_file(
            path=path,
            reuse_info=reuse_info,
            template=template,
            template_is_commented=commented,
            style=style,
            encoding=encoding,
            newline=newline,
            force_multi=multi_line,
            skip_existing=skip_existing,
            skip_unrecognised=skip_unrecognised,
            fallback_dot_license=fallback_dot_license,
            merge_copyrights=merge_copyrights,
            replace=not no_replace,
            out=sys.stdout,
        )

    sys.exit(min(result, 1))



# SPDX-FileCopyrightText: 2017 Free Software Foundation Europe e.V. <https://fsfe.org>
# SPDX-FileCopyrightText: 2022 Florian Snow <florian@familysnow.net>
# SPDX-FileCopyrightText: 2023 DB Systel GmbH
# SPDX-FileCopyrightText: 2024 Nico Rikken <nico@nicorikken.eu>
#
# SPDX-License-Identifier: GPL-3.0-or-later

"""All linting happens here. The linting here is nothing more than reading
the reports and printing some conclusions.
"""

import json
from collections import defaultdict
from io import StringIO
from itertools import chain
from pathlib import Path
from textwrap import TextWrapper
from typing import Any

from . import __REUSE_version__
from .i18n import _
from .report import ProjectReport, ProjectReportSubsetProtocol


# pylint: disable=too-many-branches,too-many-statements,too-many-locals
def format_plain(report: ProjectReport) -> str:
    """Formats data dictionary as plaintext string to be printed to sys.stdout

    Args:
        report: ProjectReport data

    Returns:
        String (in plaintext) that can be output to sys.stdout
    """
    output = StringIO()

    if not report.is_compliant:
        # Bad licenses
        if report.bad_licenses:
            output.write("# " + _("BAD LICENSES") + "\n\n")
            output.write(
                _("The following licenses are not valid SPDX licenses:") + "\n"
            )
            for path in sorted(report.bad_licenses.values()):
                output.write(f"* {path}\n")
            output.write("\n")

        # Deprecated licenses
        if report.deprecated_licenses:
            output.write("# " + _("DEPRECATED LICENSES") + "\n\n")
            output.write(
                _("The following licenses are deprecated by SPDX:") + "\n"
            )
            for lic in sorted(report.deprecated_licenses):
                output.write(f"* {lic}\n")
            output.write("\n")

        # Licenses without extension
        if report.licenses_without_extension:
            output.write("# " + _("LICENSES WITHOUT FILE EXTENSION") + "\n\n")
            output.write(
                _("The following licenses have no file extension:") + "\n"
            )
            for lic in sorted(report.licenses_without_extension):
                output.write(f"* {lic}\n")
            output.write("\n")

        # Missing licenses
        if report.missing_licenses:
            output.write("# " + _("MISSING LICENSES") + "\n\n")
            for lic, files in sorted(report.missing_licenses.items()):
                output.write(_("'{}' found in:").format(lic) + "\n")
                for file in sorted(files):
                    output.write(f"* {file}\n")
            output.write("\n")

        # Unused licenses
        if report.unused_licenses:
            output.write("# " + _("UNUSED LICENSES") + "\n\n")
            output.write(_("The following licenses are not used:") + "\n")
            for lic in sorted(report.unused_licenses):
                output.write(f"* {lic}\n")
            output.write("\n")

        # Read errors
        if report.read_errors:
            output.write("# " + _("READ ERRORS") + "\n\n")
            output.write(_("Could not read:") + "\n")
            for path in sorted(report.read_errors):
                output.write(f"* {path}\n")
            output.write("\n")

        if report.invalid_spdx_expressions:
            output.write("# " + _("INVALID SPDX LICENSE EXPRESSIONS") + "\n\n")
            for path, expressions in sorted(
                report.invalid_spdx_expressions.items()
            ):
                output.write(
                    _("'{}' contains invalid SPDX License Expressions:").format(
                        path
                    )
                    + "\n"
                )
                for expression in sorted(expressions):
                    output.write(f"* {expression}\n")
            output.write("\n")

        # Missing copyright and licensing information
        files_without_both = report.files_without_copyright.intersection(
            report.files_without_licenses
        )
        files_without_copyright_excl = (
            report.files_without_copyright - files_without_both
        )
        files_without_licenses_excl = (
            report.files_without_licenses - files_without_both
        )
        files_without_either = files_without_copyright_excl.union(
            files_without_licenses_excl
        )

        if files_without_either or files_without_both:
            header = (
                "# " + _("MISSING COPYRIGHT AND LICENSING INFORMATION") + "\n\n"
            )
            output.write(header)
        if files_without_both:
            output.write(
                _(
                    "The following files have no copyright and licensing "
                    "information:"
                )
            )
            output.write("\n")
            for file in sorted(files_without_both):
                output.write(f"* {file}\n")
            output.write("\n")

        if files_without_copyright_excl:
            output.write(
                _("The following files have no copyright information:")
            )
            output.write("\n")
            for file in sorted(files_without_copyright_excl):
                output.write(f"* {file}\n")
            output.write("\n")

        if files_without_licenses_excl:
            output.write(
                _("The following files have no licensing information:")
            )
            output.write("\n")
            for file in sorted(files_without_licenses_excl):
                output.write(f"* {file}\n")
            output.write("\n")

    output.write("# " + _("SUMMARY"))
    output.write("\n\n")

    total_files = len(report.file_reports)
    summary_contents = {
        _("Bad licenses:"): ", ".join(sorted(report.bad_licenses)),
        _("Deprecated licenses:"): ", ".join(
            sorted(report.deprecated_licenses)
        ),
        _("Licenses without file extension:"): ", ".join(
            sorted(report.licenses_without_extension)
        ),
        _("Missing licenses:"): ", ".join(sorted(report.missing_licenses)),
        _("Unused licenses:"): ", ".join(sorted(report.unused_licenses)),
        _("Used licenses:"): ", ".join(sorted(report.used_licenses)),
        _("Read errors:"): str(len(report.read_errors)),
        _("Invalid SPDX License Expressions:"): str(
            len(
                list(
                    chain.from_iterable(
                        report.invalid_spdx_expressions.values()
                    )
                )
            )
        ),
        _(
            "Files with copyright information:"
        ): f"{total_files - len(report.files_without_copyright)}"
        f" / {total_files}",
        _(
            "Files with license information:"
        ): f"{total_files - len(report.files_without_licenses)}"
        f" / {total_files}",
    }

    # Replace empty values with 0.
    summary_contents = {
        key: value if value else "0" for key, value in summary_contents.items()
    }

    for key, value in summary_contents.items():
        output.write(f"* {key} {value}\n")

    output.write("\n")
    if report.is_compliant:
        output.write(
            _(
                "Congratulations! Your project is compliant with version"
                " {} of the REUSE Specification :-)"
            ).format(__REUSE_version__)
        )
    else:
        output.write(
            _(
                "Unfortunately, your project is not compliant with version "
                "{} of the REUSE Specification :-("
            ).format(__REUSE_version__)
        )

        # Write recommendations in a nicely wrapped format
        output.write("\n\n\n# ")
        output.write(_("RECOMMENDATIONS"))
        output.write("\n\n")

        wrapper = TextWrapper(
            width=80,
            drop_whitespace=True,
            break_long_words=False,
            initial_indent="* ",
            subsequent_indent="  ",
        )
        for help_text in report.recommendations:
            output.write("\n".join(wrapper.wrap(help_text)))
            output.write("\n")

    output.write("\n")

    return output.getvalue()


def format_json(report: ProjectReport) -> str:
    """Formats data dictionary as JSON string ready to be printed to sys.stdout

    Args:
        report: Dictionary containing formatted ProjectReport data

    Returns:
        String (representing JSON) that can be output to sys.stdout
    """

    def custom_serializer(obj: Any) -> Any:
        """Custom serializer for the dictionary output of ProjectReport

        Args:
            obj: Object to be serialized
        """
        if isinstance(obj, Path):
            return str(obj)
        if isinstance(obj, set):
            return list(obj)
        raise TypeError(
            f"Object of type {obj.__class__.__name__} is not JSON serializable"
        )

    return json.dumps(
        report.to_dict_lint(),
        indent=2,
        # Serialize sets to lists
        default=custom_serializer,
    )


def _output_lines_dict(format_dict: dict[str, list[str]]) -> str:
    output = StringIO()
    for path, items in sorted(format_dict.items()):
        for item in items:
            output.write(f"{path}: {item}\n")
    return output.getvalue()


def _format_lines_subset_dict(
    report: ProjectReportSubsetProtocol,
) -> defaultdict[str, list[str]]:
    result_dict = defaultdict(list)

    # Missing licenses
    for lic, files in sorted(report.missing_licenses.items()):
        for path in files:
            result_dict[str(path)].append(
                _("missing license '{lic}'").format(lic=lic)
            )

    # Read errors
    for path in report.read_errors:
        result_dict[str(path)].append(_("read error").format(path=path))

    for path, expressions in report.invalid_spdx_expressions.items():
        for expression in sorted(expressions):
            result_dict[str(path)].append(
                _("invalid SPDX License Expression '{expression}'").format(
                    expression=expression
                )
            )

    # Without licenses
    for path in report.files_without_licenses:
        result_dict[str(path)].append(_("no license identifier"))

    # Without copyright
    for path in report.files_without_copyright:
        result_dict[str(path)].append(_("no copyright notice"))

    return result_dict


def format_lines_subset(report: ProjectReportSubsetProtocol) -> str:
    """Formats a subset of a report, namely missing licenses, read errors,
    invalid SPDX License Expressions, files without licenses, and files without
    copyright.

    Args:
        report: A populated report.

    Returns:
        String (in plaintext) that can be output to sys.stdout
    """
    return _output_lines_dict(_format_lines_subset_dict(report))


def format_lines(report: ProjectReport) -> str:
    """Formats report as plaintext strings to be printed to sys.stdout. Sorting
    of output is not guaranteed.

    Args:
        report: A populated report.

    Returns:
        String (in plaintext) that can be output to sys.stdout
    """

    def license_path(lic: str) -> str:
        """Resolve a license identifier to a license path."""
        return str(report.licenses.get(lic))

    result_dict: defaultdict[str, list[str]] = defaultdict(list)

    # Bad licenses
    for lic, path in report.bad_licenses.items():
        result_dict[str(path)].append(_("bad license '{lic}'").format(lic=lic))

    # Deprecated licenses
    for lic in report.deprecated_licenses:
        lic_path = license_path(lic)
        result_dict[lic_path].append(_("deprecated license"))

    # Licenses without extension
    for lic in report.licenses_without_extension:
        lic_path = license_path(lic)
        result_dict[lic_path].append(_("license without file extension"))

    # Unused licenses
    for lic in report.unused_licenses:
        lic_path = license_path(lic)
        result_dict[lic_path].append(_("unused license"))

    subset_dict = _format_lines_subset_dict(report)
    for key, value in subset_dict.items():
        result_dict[key].extend(value)

    return _output_lines_dict(result_dict)
